"""Extract SNAKES docstrings and generate API documentation from them
in Markdown format (one file for each Python file). This is both a
Python module and a command-line tool:

    :::console
    $ python -m snakes.utils.apidoc
    [error] Usage: python -m snakes.utils.apidoc TARGET [EXCLUDE...]
            TARGET   target directory to write files
            EXCLUDE  pattern to exclude modules (not file names)
    $ python -m snakes.utils.apidoc api snakes.lang.*
    [info] snakes -> 'api/index.md'
    ...

With respect to existing API extractors, this one takes into account
the structure of SNAKES, in particular, documentation for plugins is
correctly searched within functions `extend` and rendered accordingly.
Moreover, `apidoc` considers all the strings within modules or classes
as docstrings, which is particularly useful to document modules.

There is no real user manual for `apidoc`, in particular about its
syntax and usage, the best source of information is probably to look
at SNAKES source code. However, let's note a few points:

  * most Epydoc fields are supported: `@param`, `@type`, `@note`, ...
  * directive `# apidoc skip` allow not to include the next object in
    the generated documentation
  * directive `# apidoc stop` allow to stop the processing of current
    object (module or class)
  * directive `# apidoc include 'filename' lang='language'` allow to
    include the content of a file as a block of source code in the
    specified language
  * when statement `pass` is found in a doctest, the rest of the
    doctest is skipped, this is useful when this doctest is for test
    purpose only but does not provide useful documentation in itself
  * the syntax is that supported by Markdown Python module, with no
    extensions other that Epydoc fields
  * `apidoc` is intended to be simple and so does not support many
    options nor customisation, however, it is quite flexible already
    by the way it handles documentation

@note: in the example above we do not build doc for the sub-modules of
    `snakes.lang` because some of them are programmed for distinct
    versions of Python with incompatible syntaxes. So `apidoc` will
    fail for sure on one or another of these modules. Fortunately,
    most of the documentation for `snakes.lang` is in the package
    itself.
@note: for better rendering, `apidoc` uses [Python
    Markdown](http://pythonhosted.org/Markdown), however, it will
    handle situations when it is not installed
@warning: `apidoc` has been tested _only_ with SNAKES source, I guess
    it should work in other cases but this may required some work
"""

import sys, os, os.path
import inspect, fnmatch, collections, re, shlex
import textwrap, doctest
import snakes
from snakes.lang import unparse
from snakes.lang.python.parser import parse, ast

try :
    import markdown
except :
    markdown = None

##
## console messages
##

CLEAR = "\033[0m"
BLUE = "\033[1;34m"
BOLD = "\033[1;38m"
GRAY = "\033[1;30m"
LIGHTGRAY = "\033[0;30m"
GREEN = "\033[1;32m"
CYAN = "\033[1;36m"
MAGENTA = "\033[1;35m"
RED = "\033[1;31m"
YELLOW = "\033[1;33m"

# apidoc skip
def log (message, color=None, eol=True) :
    if color :
        sys.stderr.write(color)
    if isinstance(message, (list, tuple)) :
        message = " ".join(str(m) for m in message)
    sys.stderr.write(message)
    if color :
        sys.stderr.write(CLEAR)
    if eol :
        sys.stderr.write("\n")
    sys.stderr.flush()

# apidoc skip
def info (message, eol=True) :
    log("[info] ", BLUE, False)
    log(message, eol=eol)

# apidoc skip
def warn (message, eol=True) :
    log("[warn] ", YELLOW, False)
    log(message, eol=eol)

# apidoc skip
def err (message, eol=True) :
    log("[error] ", RED, False)
    log(message, eol=eol)

# apidoc skip
def die (message, code=1) :
    err(message)
    sys.exit(code)

##
## extract doc
##

class DocExtract (object) :
    """The class that extracts documentation and renders it
    """
    def __init__ (self, inpath, outpath, exclude=[]) :
        """
        @param inpath: directory where the source code is searched for
        @type inpath: `str`
        @param outpath: directory where the documentation is generated
        @type outpath: `str`
        @param exclude: a list of glob patterns to exclude modules
            (not file names, but Python modules names)
        @type exclude: `list`
        """
        self.path = inpath.rstrip(os.sep)
        self.outpath = outpath
        self.out = None
        self.exclude = exclude
        self._last = "\n\n"
    def md (self, text, inline=True) :
        """Return the Markdow rendering of `text`, include `<p>` if
        `inline` is `False`. If Python Markdown module is not
        installed, `text` is returned unchanged.

        @param text: the text to be rendered
        @type text: `str`
        @param inline: `True` if this text is part of a larger
            paragraph, `False` if this is a paragraph by itself
        @type inline: `bool`
        @return: HTML text
        @rtype: `str`
        """
        if markdown is None :
            return text
        elif inline :
            return re.sub("</?p>", "\n", markdown.markdown(text), re.I)
        else :
            return markdown.markdown(text)
    def openout (self, path) :
        """Open in `self.out` the output file where the conversion of
        input file `path` will be rendered. If `self.out` is already
        opened, it is first closed.

        Return a Boolean indicating if `path` should be ignored, in
        which case the output file is not opened. This occurs either
        if `path` correspond to a module that has been excluded or if
        the output file exists and is newer than the input file.

        @param path: input file to be converted
        @type path: `str`
        @return: `True` is output file is open and ready, `False` if
            input file should be skipped
        @rtype: `bool`
        @note: package files called `__init__.py` are converted to
            files called `index.md` so that when converted to HTML,
            this will yield a file called `index.html`
        """
        if self.out is not None :
            self.out.close()
        relpath = path[len(os.path.dirname(self.path)):].strip(os.sep)
        parts = relpath.split(os.sep)
        relpath = os.path.dirname(relpath.split(os.sep, 1)[-1])
        self.package = (parts[-1] == "__init__.py")
        if self.package :
            self.module = ".".join(parts[:-1])
            target = "index.md"
        else :
            parts[-1] = os.path.splitext(parts[-1])[0]
            self.module = ".".join(parts)
            target =  parts[-1] + ".md"
        if any(fnmatch.fnmatch(self.module, glob) for glob in self.exclude) :
            warn("skip %s" % self.module)
            return False
        outdir = os.path.join(self.outpath, relpath)
        outpath = os.path.join(outdir, target)
        if os.path.exists(outpath) :
            if os.stat(path).st_mtime <= os.stat(outpath).st_mtime :
                return False
        self.inpath = path
        info("%s -> %r" % (self.module, outpath))
        if not os.path.exists(outdir) :
            os.makedirs(outdir)
        self.out = open(outpath, "w")
        self.classname = None
        return True
    def write (self, text) :
        """Write `text` to output file

        @param text: the text to write
        @type text: `str`
        """
        if len(text) > 1 :
            self._last = text[-2:]
        elif len(text) > 0 :
            self._last = self._last[-1] + text[-1]
        else :
            return
        self.out.write(text)
    def newline (self) :
        """Write a blank line to output file, never more than one
        blank line is written so there is no need to be carefull about
        how many successive times `newline` is called (which would be
        complicated when multiple calls are made from different
        locations).
        """
        if self._last != "\n\n" :
            self.write("\n")
    def writeline (self, text="") :
        """Write a line of text to output file, ensuring there is a
        end-of-line character at the end (and removing those that may
        exist in `text`).

        @param text: the text to write
        @type text: `str`
        """
        self.write(text.rstrip() + "\n")
    def writetext (self, text, **args) :
        """Write some `text` to output file, with line wrapping.

        @param text: the text to write
        @type text: `str`
        @param args: additional arguments to `textwrap.wrap`
        """
        br = args.pop("break_on_hyphens", False)
        for line in textwrap.wrap(text, break_on_hyphens=br, **args) :
            self.writeline(line)
    def writelist (self, text, bullet="  * ", **args) :
        """Write one list item to output file, wrapping the text
        properly.

        @param text: the text within the item
        @type text: `str`
        @param bullet: list marker
        @type bullet: `str`
        @param args: additional arguments to `textwrap.wrap`
        """
        br = args.pop("break_on_hyphens", False)
        for line in textwrap.wrap(text, break_on_hyphens=br,
                                  initial_indent=bullet,
                                  subsequent_indent=" "*len(bullet),
                                  **args) :
            self.writeline(line)
    def process (self) :
        """Main method that process input directory and write
        converted files to output directory.

        Each input file is parsed using `snakes.lang.python` to an AST
        that is then traversed for processing.
        """
        for dirpath, dirnames, filenames in os.walk(self.path) :
            for name in sorted(filenames) :
                if not name.endswith(".py") or name.startswith(".") :
                    continue
                path = os.path.join(dirpath, name)
                if not self.openout(path) :
                    continue
                node = parse(open(path).read())
                if ".plugins." in self.module :
                    self.visit_plugin(node)
                else :
                    self.visit_module(node)
    def _pass (self, node) :
        pass
    def directive (self, node) :
        """Return directives (skip, stop, include) for this node, or
        `None`

        @param node: an AST node
        @type node: `AST`
        """
        lines = node.st.text.lexer.lines
        num = node.st.srow - 2
        while num >= 0 and (not lines[num].strip()
                            or lines[num].strip().startswith("@")) :
            num -= 1
        if num >= 0 and lines[num].strip().startswith("#") :
            dirline = lines[num].lstrip("# \t").rstrip()
            items = shlex.split(dirline)
            if len(items) >= 2 and items[0].lower() == "apidoc" :
                if len(items) == 2 and items[1].lower() in ("skip", "stop") :
                    return items[1]
                elif items[1].lower() == "include" :
                    path = items[2]
                    try :
                        args = dict(i.split("=", 1) for i in items[3:])
                    except :
                        err("invalid directive %r (line %s)"
                            % (dirline, num+1))
                    self.write_include(path, **args)
                else :
                    err("unknown directive %r (line %s)"
                        % (items[1], num+1))
                    return None
    def children (self, node) :
        """Iterates over the children of `node`

        @param node: an AST node
        @type node: `AST`
        @return: all the children but the `skip`ped ones, until the
            end of `stop` directive is found
        @rtype: `generator`
        """
        for child in ast.iter_child_nodes(node) :
            directive = self.directive(child)
            if directive == "skip" :
                continue
            elif directive == "stop" :
                break
            yield child
    def visit (self, node) :
        """Generic visit of a node that actually dispatch to the
        appropriate method `visit_...`
        """
        name = getattr(node, "name", "__")
        if (name.startswith("_") and not (name.startswith("__")
                                          and name.endswith("__"))) :
            return
        try :
            getattr(self, "visit_" + node.__class__.__name__,
                    self._pass)(node)
        except :
            src = node.st.source()
            if len(src) > 40 :
                src = src[:40] + "..."
            err("line %s source %r" % (node.lineno, src))
            raise
    def visit_module (self, node) :
        """Visit a node that is a module
        """
        self.write_module()
        for child in self.children(node) :
            self.visit(child)
    def visit_plugin (self, node) :
        """Visit a node that is a plugin
        """
        self.write_module()
        extend = None
        for child in self.children(node) :
            if (getattr(child, "name", None) == "extend"
                and isinstance(child, ast.FunctionDef)) :
                extend = child
            else :
                self.visit(child)
        self.write_plugin()
        for child in self.children(extend) :
            self.visit(child)
    def visit_ClassDef (self, node) :
        """Visit a node that is a class definition
        """
        self.write_class(node)
        self.classname = node.name
        for child in self.children(node) :
            self.visit(child)
        self.classname = None
    def visit_FunctionDef (self, node) :
        """Visit a node that is a function definition
        """
        self.write_function(node)
        self.args = [n.arg for n in node.args.args]
        if self.args and self.args[0] == "self" :
            del self.args[0]
        if node.args.vararg :
            self.args.append(node.args.vararg)
        if node.args.kwarg :
            self.args.append(node.args.kwarg)
        self.visit(node.body[0])
        self.args = []
    def visit_Expr (self, node) :
        """Visit a node that is an expression
        """
        self.visit(node.value)
    def visit_Str (self, node) :
        """Visit a node that is a string literal
        """
        self.write_doc(inspect.cleandoc(node.s))
    def write_module (self) :
        """Write the documentation about a module (not its content),
        which is just a title. The module name is in `self.module` and
        there is also a Boolean `self.package` to indicate if this
        module is a actually package.
        """
        if self.package :
            self.writeline("# Package `%s` #" % self.module)
        else :
            self.writeline("# Module `%s` #" % self.module)
        self.newline()
    def write_plugin (self) :
        """Write the documentation about function `extend` in a plugin
        (not its content), which is just a title
        """
        self.writeline("## Extensions ##")
        self.newline()
    def write_function (self, node) :
        """Write the documentation about a function or method (which
        can be differentiated by looking if `self.classname` is not
        None)
        """
        self.newline()
        if self.classname :
            self.writeline("#### Method `%s.%s` ####"
                           % (self.classname, node.name))
        else :
            self.writeline("### Function `%s` ###" % node.name)
        self.newline()
        self.write_def(node)
    def write_class (self, node) :
        """Write the documentation about a class
        """
        self.newline()
        self.writeline("### Class `%s` ###" % node.name)
        self.write_def(node)
    def write_def (self, node) :
        """Write the documentation about a function or method
        definition (parameters, decorators, etc.)
        """
        indent = node.st.scol
        if node.st[0].symbol == "decorated" :
            srow, scol = node.st[0][0][0][0].srow, node.st[0][0][0][0].scol
            erow, ecol = node.st[0][1][-2].erow, node.st[0][1][-2].ecol
        else :
            srow, scol = node.st[0].srow, node.st[0].scol
            erow, ecol = node.st[0][-2].erow, node.st[0][-2].ecol
        lines = node.st.text.lexer.lines
        if srow == erow :
            source = [lines[srow-1][scol:ecol+1]]
        else :
            source = lines[srow-1:erow]
            source[0] = source[0][scol:]
            source[1:-1] = [s[indent:] for s in source[1:-1]]
            source[-1] = source[-1][indent:ecol+1].rstrip() + " ..."
        self.newline()
        self.writeline("    :::python")
        for line in source :
            self.writeline("    " + line)
        self.newline()
    parse = doctest.DocTestParser().parse
    def write_doc (self, doc) :
        """Write the content of a docstring that is parsed to extract
        doctests, Epydoc fields and plain text
        """
        if doc is None :
            return
        docs = self.parse(doc)
        test, skip = False, False
        for doc in docs :
            if isinstance(doc, str) :
                doc = doc.strip()
                if test :
                    if not doc :
                        continue
                    test, skip = False, False
                self.newline()
                lines = doc.strip().splitlines()
                for num, line in enumerate(lines) :
                    if line.startswith("@") :
                        self.write_epydoc("\n".join(lines[num:]))
                        break
                    self.writeline(line)
            elif not skip :
                if doc.source.strip() == "pass" :
                    skip = True
                else :
                    if not test :
                        test = True
                        self.newline()
                        self.writetext("<!-- this comment avoids a bug in"
                                       " Markdown parsing -->")
                        self.writeline("    :::pycon")
                    for i, line in enumerate(doc.source.splitlines()) :
                        if i > 0 :
                            self.writeline("    ... %s" % line)
                        else :
                            self.writeline("    >>> %s" % line)
                    for line in doc.want.splitlines() :
                        self.writeline("    %s" % line)
    def write_epydoc (self, doc) :
        """Write a block of epydoc fields
        """
        info = {"param" : {},
                "type" : {},
                "keyword" : {},
                "raise" : {},
                "todo" : [],
                "note" : [],
                "attention": [],
                "bug" : [],
                "warning" : [],
                }
        for item in doc.lstrip("@").split("\n@") :
            left, text = item.split(":", 1)
            left = left.split()
            assert 1 <= len(left) <= 2, "unsupported item %r" % item
            if len(left) == 1 :
                left.append(None)
            tag, name = [x.strip() if x else x for x in left]
            text = " ".join(text.strip().split())
            if isinstance(info.get(tag, None), list) :
                assert name is None, "unsupported item %r" % item
                info[tag].append(text)
            elif isinstance(info.get(tag, None), dict) :
                assert name is not None, "unsupported item %r" % item
                assert name not in info[tag], "duplicated item %r" % item
                info[tag][name] = text
            else :
                assert name is None, "unsupported item %r" % item
                assert tag not in info, "duplicated tag %r" % item
                info[tag] = text
        if any(k in info for k in ("author", "organization", "copyright",
                                   "license", "contact")) :
            self.newline()
            self.writeline('<ul id="api-info">')
            for tag in ("author", "organization", "contact",
                        "copyright", "license", ) :
                if tag in info :
                    self.writeline('<li id="api-%s">' % tag)
                    self.writetext('<span class="api-title">%s:</span> %s'
                                   % (tag.capitalize(),
                                      self.md(info[tag])),
                                   subsequent_indent="  ")
                    self.writeline('</li>')
            self.writeline('</ul>')
        if any(info[k] for k in
               ("todo", "note", "attention", "bug", "warning")) :
            self.newline()
            self.writeline('<div id="api-remarks">')
            for tag in ("note", "todo", "attention", "bug", "warning") :
                for text in info[tag] :
                    self.writeline('<div class="api-%s">' % tag)
                    self.writetext('<span class="api-title">%s:</span> %s'
                                   % (tag.capitalize(), self.md(text)),
                                   subsequent_indent="  ")
                    self.writeline('</div>')
            self.writeline('</div>')
        if (any(info[k] for k in ("param", "type", "keyword"))
            or any(k in info for k in ("return", "rtype"))) :
            self.newline()
            self.writeline("##### Call API #####")
            self.newline()
            for arg in self.args :
                if arg in info["param"] :
                    self.writelist("`%s %s`: %s"
                                   % (info["type"].get(arg, "object").strip("`"),
                                      arg,
                                      info["param"][arg]))
                else :
                    self.writelist("`%s %s`"
                                   % (info["type"].get(arg, "object").strip("`"),
                                      arg))
            for kw, text in sorted(info["keyword"].items()) :
                self.writelist("keyword `%s`: %s" % (kw, text))
            if any(k in info for k in ("return", "rtype")) :
                if "return" in info :
                    self.writelist("`return %s`: %s"
                                   % (info.get("rtype", "object").strip("`"),
                                      info["return"]))
                else :
                    self.writelist("`return %s`"
                                   % (info.get("rtype", "object").strip("`")))
        if info["raise"] :
            self.newline()
            self.writeline("##### Exceptions #####")
            self.newline()
            for exc, reason in sorted(info["raise"].items()) :
                self.writelist("`%s`: %s" % (exc, reason))
            self.newline()
    def write_include (self, name, lang="python", first=1, last=-1) :
        """Write a block of source code included through directive
        `apidoc include ...`
        """
        first, last = int(first), int(last)
        if os.path.exists(name) :
            path = name
        else :
            path = os.path.join(os.path.dirname(self.inpath), name)
        if not os.path.exists(path) :
            err("include file %r not found" % name)
        with open(path) as infile :
            self.newline()
            self.writeline("    :::%s" % lang)
            for i, line in enumerate(infile) :
                if first <= i+1 and (last == -1 or i+1 <= last) :
                    self.writeline("    " + line.rstrip())
            self.newline()

def main (finder, args, source=None) :
    """Main function of `apidoc`. Source directory may be given
    explicitly, otherwise it is computed as follows:

        :::python
        import snakes
        source = os.path.dirname(snakes.__file__)

    So it is important to set `PYTHONPATH` or `sys.path` if you want
    to use a non-standard location for SNAKES.

    @param finder: `DocExtract` or a subclass of it to perform the
        processing
    @type finder: `class`
    @param args: command line arguments
    @type args: `list`
    @param source: the source directory for the code to be documented,
        or `None` to extract documentation from SNAKES
    @type source: `str`
    """
    try :
        if source is None :
            source = os.path.dirname(snakes.__file__)
        target = args[0]
        exclude = args[1:]
        if not os.path.isdir(source) :
            raise Exception("could not find SNAKES sources")
        elif not os.path.isdir(target) :
            raise Exception("no directory %r" % target)
    except (ValueError, IndexError) :
        die("""Usage: python -m snakes.utils.apidoc TARGET [EXCLUDE...]
        TARGET   target directory to write files
        EXCLUDE  pattern to exclude modules (not file names)""")
    except Exception as error :
        die(str(error))
    finder(source, target, exclude).process()

if __name__ == "__main__" :
    main(DocExtract, sys.argv[1:])



import sys

import snakes.plugins
snakes.plugins.load("gv", "snakes.nets", "snk")
from snk import *

n = loads(sys.argv[1])
g = StateGraph(n)
for s in g :
    m = g.net.get_marking()
    # safety property: train present => gates closed
    if ("train().crossing" in m
        and True in m["train().crossing"]
        and "closed" not in m["gate().state"]) :
        print("%s %s" % (s, m))
print("checked %s states" % len(g))

g.draw(sys.argv[1].rsplit(".", 1)[0] + "-states.png")



# this file has been automatically generated running:
# snakes/lang/asdl.py --output=snakes/lang/ctlstar/asdl.py snakes/lang/ctlstar/ctlstar.asdl
# timestamp: 2011-11-16 13:30:34.219385

from snakes.lang import ast
from ast import *

class _AST (ast.AST):
    def __init__ (self, **ARGS):
        ast.AST.__init__(self)
        for k, v in ARGS.items():
            setattr(self, k, v)

class expr_context (_AST):
    pass

class Load (expr_context):
    _fields = ()
    _attributes = ()

class Store (expr_context):
    _fields = ()
    _attributes = ()

class Del (expr_context):
    _fields = ()
    _attributes = ()

class AugLoad (expr_context):
    _fields = ()
    _attributes = ()

class AugStore (expr_context):
    _fields = ()
    _attributes = ()

class Param (expr_context):
    _fields = ()
    _attributes = ()

class comprehension (_AST):
    _fields = ('target', 'iter', 'ifs')
    _attributes = ()
    def __init__ (self, target, iter, ifs=[], **ARGS):
        _AST.__init__(self, **ARGS)
        self.target = target
        self.iter = iter
        self.ifs = list(ifs)

class arg (_AST):
    _fields = ('arg', 'annotation')
    _attributes = ()
    def __init__ (self, arg, annotation=None, **ARGS):
        _AST.__init__(self, **ARGS)
        self.arg = arg
        self.annotation = annotation

class operator (_AST):
    pass

class Add (operator):
    _fields = ()
    _attributes = ()

class Sub (operator):
    _fields = ()
    _attributes = ()

class Mult (operator):
    _fields = ()
    _attributes = ()

class Div (operator):
    _fields = ()
    _attributes = ()

class Mod (operator):
    _fields = ()
    _attributes = ()

class Pow (operator):
    _fields = ()
    _attributes = ()

class LShift (operator):
    _fields = ()
    _attributes = ()

class RShift (operator):
    _fields = ()
    _attributes = ()

class BitOr (operator):
    _fields = ()
    _attributes = ()

class BitXor (operator):
    _fields = ()
    _attributes = ()

class BitAnd (operator):
    _fields = ()
    _attributes = ()

class FloorDiv (operator):
    _fields = ()
    _attributes = ()

class slice (_AST):
    pass

class Slice (slice):
    _fields = ('lower', 'upper', 'step')
    _attributes = ()
    def __init__ (self, lower=None, upper=None, step=None, **ARGS):
        slice.__init__(self, **ARGS)
        self.lower = lower
        self.upper = upper
        self.step = step

class ExtSlice (slice):
    _fields = ('dims',)
    _attributes = ()
    def __init__ (self, dims=[], **ARGS):
        slice.__init__(self, **ARGS)
        self.dims = list(dims)

class Index (slice):
    _fields = ('value',)
    _attributes = ()
    def __init__ (self, value, **ARGS):
        slice.__init__(self, **ARGS)
        self.value = value

class excepthandler (_AST):
    pass

class ExceptHandler (excepthandler):
    _fields = ('type', 'name', 'body')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, type=None, name=None, body=[], lineno=0, col_offset=0, **ARGS):
        excepthandler.__init__(self, **ARGS)
        self.type = type
        self.name = name
        self.body = list(body)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class arguments (_AST):
    _fields = ('args', 'vararg', 'varargannotation', 'kwonlyargs', 'kwarg', 'kwargannotation', 'defaults', 'kw_defaults')
    _attributes = ()
    def __init__ (self, args=[], vararg=None, varargannotation=None, kwonlyargs=[], kwarg=None, kwargannotation=None, defaults=[], kw_defaults=[], **ARGS):
        _AST.__init__(self, **ARGS)
        self.args = list(args)
        self.vararg = vararg
        self.varargannotation = varargannotation
        self.kwonlyargs = list(kwonlyargs)
        self.kwarg = kwarg
        self.kwargannotation = kwargannotation
        self.defaults = list(defaults)
        self.kw_defaults = list(kw_defaults)

class ctlbinary (_AST):
    pass

class boolop (ctlbinary):
    _fields = ()
    _attributes = ()

class Imply (ctlbinary):
    _fields = ()
    _attributes = ()

class Iff (ctlbinary):
    _fields = ()
    _attributes = ()

class Until (ctlbinary):
    _fields = ()
    _attributes = ()

class WeakUntil (ctlbinary):
    _fields = ()
    _attributes = ()

class Release (ctlbinary):
    _fields = ()
    _attributes = ()

class ctlunary (_AST):
    pass

class notop (ctlunary):
    _fields = ()
    _attributes = ()

class All (ctlunary):
    _fields = ()
    _attributes = ()

class Exists (ctlunary):
    _fields = ()
    _attributes = ()

class Next (ctlunary):
    _fields = ()
    _attributes = ()

class Future (ctlunary):
    _fields = ()
    _attributes = ()

class Globally (ctlunary):
    _fields = ()
    _attributes = ()

class form (_AST):
    pass

class atom (form):
    _fields = ()
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, lineno=0, col_offset=0, **ARGS):
        form.__init__(self, **ARGS)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class CtlUnary (form):
    _fields = ('op', 'child')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, op, child, lineno=0, col_offset=0, **ARGS):
        form.__init__(self, **ARGS)
        self.op = op
        self.child = child
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class CtlBinary (form):
    _fields = ('op', 'left', 'right')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, op, left, right, lineno=0, col_offset=0, **ARGS):
        form.__init__(self, **ARGS)
        self.op = op
        self.left = left
        self.right = right
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class unaryop (_AST):
    pass

class Invert (unaryop):
    _fields = ()
    _attributes = ()

class notop (unaryop):
    _fields = ()
    _attributes = ()

class UAdd (unaryop):
    _fields = ()
    _attributes = ()

class USub (unaryop):
    _fields = ()
    _attributes = ()

class boolop (_AST):
    pass

class And (boolop):
    _fields = ()
    _attributes = ()

class Or (boolop):
    _fields = ()
    _attributes = ()

class stmt (_AST):
    pass

class FunctionDef (stmt):
    _fields = ('name', 'args', 'body', 'decorator_list', 'returns')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, args, body=[], decorator_list=[], returns=None, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.name = name
        self.args = args
        self.body = list(body)
        self.decorator_list = list(decorator_list)
        self.returns = returns
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class ClassDef (stmt):
    _fields = ('name', 'bases', 'keywords', 'starargs', 'kwargs', 'body', 'decorator_list')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, bases=[], keywords=[], starargs=None, kwargs=None, body=[], decorator_list=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.name = name
        self.bases = list(bases)
        self.keywords = list(keywords)
        self.starargs = starargs
        self.kwargs = kwargs
        self.body = list(body)
        self.decorator_list = list(decorator_list)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Return (stmt):
    _fields = ('value',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value=None, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.value = value
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Delete (stmt):
    _fields = ('targets',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, targets=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.targets = list(targets)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Assign (stmt):
    _fields = ('targets', 'value')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value, targets=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.targets = list(targets)
        self.value = value
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class AugAssign (stmt):
    _fields = ('target', 'op', 'value')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, target, op, value, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.target = target
        self.op = op
        self.value = value
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class For (stmt):
    _fields = ('target', 'iter', 'body', 'orelse')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, target, iter, body=[], orelse=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.target = target
        self.iter = iter
        self.body = list(body)
        self.orelse = list(orelse)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class While (stmt):
    _fields = ('test', 'body', 'orelse')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, test, body=[], orelse=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.test = test
        self.body = list(body)
        self.orelse = list(orelse)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class If (stmt):
    _fields = ('test', 'body', 'orelse')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, test, body=[], orelse=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.test = test
        self.body = list(body)
        self.orelse = list(orelse)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class With (stmt):
    _fields = ('context_expr', 'optional_vars', 'body')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, context_expr, optional_vars=None, body=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.context_expr = context_expr
        self.optional_vars = optional_vars
        self.body = list(body)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Raise (stmt):
    _fields = ('exc', 'cause')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, exc=None, cause=None, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.exc = exc
        self.cause = cause
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class TryExcept (stmt):
    _fields = ('body', 'handlers', 'orelse')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, body=[], handlers=[], orelse=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.body = list(body)
        self.handlers = list(handlers)
        self.orelse = list(orelse)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class TryFinally (stmt):
    _fields = ('body', 'finalbody')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, body=[], finalbody=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.body = list(body)
        self.finalbody = list(finalbody)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Assert (stmt):
    _fields = ('test', 'msg')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, test, msg=None, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.test = test
        self.msg = msg
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Import (stmt):
    _fields = ('names',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, names=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.names = list(names)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class ImportFrom (stmt):
    _fields = ('module', 'names', 'level')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, module, names=[], level=None, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.module = module
        self.names = list(names)
        self.level = level
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Exec (stmt):
    _fields = ('body', 'globals', 'locals')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, body, globals=None, locals=None, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.body = body
        self.globals = globals
        self.locals = locals
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Global (stmt):
    _fields = ('names',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, names=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.names = list(names)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Nonlocal (stmt):
    _fields = ('names',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, names=[], lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.names = list(names)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Expr (stmt):
    _fields = ('value',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.value = value
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Pass (stmt):
    _fields = ()
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Break (stmt):
    _fields = ()
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Continue (stmt):
    _fields = ()
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, lineno=0, col_offset=0, **ARGS):
        stmt.__init__(self, **ARGS)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class notop (_AST):
    pass

class Not (notop):
    _fields = ()
    _attributes = ()

class ctlstar (_AST):
    pass

class Spec (ctlstar):
    _fields = ('atoms', 'properties', 'main')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, atoms=[], properties=[], main=None, lineno=0, col_offset=0, **ARGS):
        ctlstar.__init__(self, **ARGS)
        self.atoms = list(atoms)
        self.properties = list(properties)
        self.main = main
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class atom (_AST):
    pass

class InPlace (atom):
    _fields = ('data', 'place')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, place, data=[], lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.data = list(data)
        self.place = place
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class NotInPlace (atom):
    _fields = ('data', 'place')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, place, data=[], lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.data = list(data)
        self.place = place
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class EmptyPlace (atom):
    _fields = ('place',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, place, lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.place = place
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class MarkedPlace (atom):
    _fields = ('place',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, place, lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.place = place
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Deadlock (atom):
    _fields = ()
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Boolean (atom):
    _fields = ('val',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, val, lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.val = val
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Instance (atom):
    _fields = ('name', 'args')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, args=[], lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.name = name
        self.args = list(args)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Quantifier (atom):
    _fields = ('op', 'vars', 'place', 'child', 'distinct')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, op, place, child, distinct, vars=[], lineno=0, col_offset=0, **ARGS):
        atom.__init__(self, **ARGS)
        self.op = op
        self.vars = list(vars)
        self.place = place
        self.child = child
        self.distinct = distinct
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class cmpop (_AST):
    pass

class Eq (cmpop):
    _fields = ()
    _attributes = ()

class NotEq (cmpop):
    _fields = ()
    _attributes = ()

class Lt (cmpop):
    _fields = ()
    _attributes = ()

class LtE (cmpop):
    _fields = ()
    _attributes = ()

class Gt (cmpop):
    _fields = ()
    _attributes = ()

class GtE (cmpop):
    _fields = ()
    _attributes = ()

class Is (cmpop):
    _fields = ()
    _attributes = ()

class IsNot (cmpop):
    _fields = ()
    _attributes = ()

class In (cmpop):
    _fields = ()
    _attributes = ()

class NotIn (cmpop):
    _fields = ()
    _attributes = ()

class keyword (_AST):
    _fields = ('arg', 'value')
    _attributes = ()
    def __init__ (self, arg, value, **ARGS):
        _AST.__init__(self, **ARGS)
        self.arg = arg
        self.value = value

class ctlarg (_AST):
    pass

class Place (ctlarg):
    _fields = ('name', 'place')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, place, lineno=0, col_offset=0, **ARGS):
        ctlarg.__init__(self, **ARGS)
        self.name = name
        self.place = place
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Token (ctlarg):
    _fields = ('name', 'place')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, place, lineno=0, col_offset=0, **ARGS):
        ctlarg.__init__(self, **ARGS)
        self.name = name
        self.place = place
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Argument (ctlarg):
    _fields = ('name', 'value', 'type')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, value, type, lineno=0, col_offset=0, **ARGS):
        ctlarg.__init__(self, **ARGS)
        self.name = name
        self.value = value
        self.type = type
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class expr (_AST):
    pass

class BoolOp (expr):
    _fields = ('op', 'values')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, op, values=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.op = op
        self.values = list(values)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class BinOp (expr):
    _fields = ('left', 'op', 'right')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, left, op, right, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.left = left
        self.op = op
        self.right = right
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class UnaryOp (expr):
    _fields = ('op', 'operand')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, op, operand, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.op = op
        self.operand = operand
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Lambda (expr):
    _fields = ('args', 'body')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, args, body, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.args = args
        self.body = body
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class IfExp (expr):
    _fields = ('test', 'body', 'orelse')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, test, body, orelse, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.test = test
        self.body = body
        self.orelse = orelse
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Dict (expr):
    _fields = ('keys', 'values')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, keys=[], values=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.keys = list(keys)
        self.values = list(values)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Set (expr):
    _fields = ('elts',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, elts=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.elts = list(elts)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class ListComp (expr):
    _fields = ('elt', 'generators')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, elt, generators=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.elt = elt
        self.generators = list(generators)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class SetComp (expr):
    _fields = ('elt', 'generators')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, elt, generators=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.elt = elt
        self.generators = list(generators)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class DictComp (expr):
    _fields = ('key', 'value', 'generators')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, key, value, generators=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.key = key
        self.value = value
        self.generators = list(generators)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class GeneratorExp (expr):
    _fields = ('elt', 'generators')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, elt, generators=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.elt = elt
        self.generators = list(generators)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Yield (expr):
    _fields = ('value',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.value = value
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Compare (expr):
    _fields = ('left', 'ops', 'comparators')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, left, ops=[], comparators=[], lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.left = left
        self.ops = list(ops)
        self.comparators = list(comparators)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Call (expr):
    _fields = ('func', 'args', 'keywords', 'starargs', 'kwargs')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, func, args=[], keywords=[], starargs=None, kwargs=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.func = func
        self.args = list(args)
        self.keywords = list(keywords)
        self.starargs = starargs
        self.kwargs = kwargs
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Num (expr):
    _fields = ('n',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, n, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.n = n
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Str (expr):
    _fields = ('s',)
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, s, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.s = s
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Ellipsis (expr):
    _fields = ()
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Attribute (expr):
    _fields = ('value', 'attr', 'ctx')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value, attr, ctx=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.value = value
        self.attr = attr
        self.ctx = ctx
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Subscript (expr):
    _fields = ('value', 'slice', 'ctx')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value, slice, ctx=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.value = value
        self.slice = slice
        self.ctx = ctx
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Starred (expr):
    _fields = ('value', 'ctx')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, value, ctx=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.value = value
        self.ctx = ctx
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Name (expr):
    _fields = ('id', 'ctx')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, id, ctx=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.id = id
        self.ctx = ctx
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class List (expr):
    _fields = ('elts', 'ctx')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, elts=[], ctx=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.elts = list(elts)
        self.ctx = ctx
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Tuple (expr):
    _fields = ('elts', 'ctx')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, elts=[], ctx=None, lineno=0, col_offset=0, **ARGS):
        expr.__init__(self, **ARGS)
        self.elts = list(elts)
        self.ctx = ctx
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class ctldecl (_AST):
    pass

class Atom (ctldecl):
    _fields = ('name', 'args', 'params', 'body')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, args=[], params=[], body=[], lineno=0, col_offset=0, **ARGS):
        ctldecl.__init__(self, **ARGS)
        self.name = name
        self.args = list(args)
        self.params = list(params)
        self.body = list(body)
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class Property (ctldecl):
    _fields = ('name', 'args', 'params', 'body')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, body, args=[], params=[], lineno=0, col_offset=0, **ARGS):
        ctldecl.__init__(self, **ARGS)
        self.name = name
        self.args = list(args)
        self.params = list(params)
        self.body = body
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)

class alias (_AST):
    _fields = ('name', 'asname')
    _attributes = ()
    def __init__ (self, name, asname=None, **ARGS):
        _AST.__init__(self, **ARGS)
        self.name = name
        self.asname = asname

class ctlparam (_AST):
    pass

class Parameter (ctlparam):
    _fields = ('name', 'type')
    _attributes = ('lineno', 'col_offset')
    def __init__ (self, name, type, lineno=0, col_offset=0, **ARGS):
        ctlparam.__init__(self, **ARGS)
        self.name = name
        self.type = type
        self.lineno = int(lineno)
        self.col_offset = int(col_offset)



"""This module proposes hashable version of the mutable containers
`list`, `dict` and `set`, called respectively `hlist`, `hdict` and
`hset`.

After one object has been hashed, it becomes not mutable and raises a
`ValueError` if a method which changes the object (let call a
_mutation_ such a method) is invoqued. The object can be then
un-hashed by calling `unhash` on it so that it becomes mutable again.
Notice that this may cause troubles if the object is stored in a set
or dict which uses its hashcode to locate it. Notice also that
hashable containers cannot be hashed if they contain non hashable
objects.

>>> l = hlist(range(5))
>>> l
hlist([0, 1, 2, 3, 4])
>>> _ = hash(l)
>>> l.append(5)
Traceback (most recent call last):
...
ValueError: hashed 'hlist' object is not mutable
>>> unhash(l)
>>> l.append(5)
>>> l
hlist([0, 1, 2, 3, 4, 5])

Testing if a `hlist` is in a `set`, `dict`, `hset` or `hdict` causes
its hashing. If this is not desirable, you should either compensate
with a call to `unhash`, or test if a copy is in the set:

>>> s = set()
>>> s.add(list(range(4)))
Traceback (most recent call last):
...
TypeError: ...
>>> s.add(hlist(range(4)))
>>> l = hlist(range(3))
>>> l in s
False
>>> l.append(3)
Traceback (most recent call last):
...
ValueError: hashed 'hlist' object is not mutable
>>> unhash(l)
>>> l.append(3)
>>> l[:] in s
True
>>> l.append(4)
"""

import inspect
from operator import xor
from snakes.compat import *

def unhash (obj) :
    """Make the object mutable again. This should be used with
    caution, especially if the object is stored in a dict or a set.

    >>> l = hlist(range(3))
    >>> _ = hash(l)
    >>> l.append(3)
    Traceback (most recent call last):
    ...
    ValueError: hashed 'hlist' object is not mutable
    >>> unhash(l)
    >>> l.append(3)

    @param obj: any object
    @type obj: `object`
    """
    try :
        del obj._hash
    except :
        pass

# apidoc skip
def hashable (cls) :
    """Wrap methods in a class in order to make it hashable.
    """
    classname, bases, classdict = cls.__name__, cls.__bases__, cls.__dict__
    for name, attr in classdict.items() :
        try :
            doc = inspect.getdoc(attr)
            if doc is None :
                attr.__doc__ = getattr(bases[0], name).__doc__
            else :
                attr.__doc__ = "\n".join([inspect.getdoc(getattr(bases[0],
                                                                 name)),
                                          doc])
        except :
            pass
    def __hash__ (self) :
        if not hasattr(self, "_hash") :
            self._hash = self.__hash__.HASH(self)
        return self._hash
    __hash__.HASH = classdict["__hash__"]
    __hash__.__doc__ = classdict["__hash__"].__doc__
    cls.__hash__ = __hash__
    def __mutable__ (self) :
        "Raise `ValueError` if the %s has been hashed."
        if self.hashed() :
            raise ValueError("hashed '%s' object is not mutable" % classname)
    try :
        __mutable__.__doc__ = __mutable__.__doc__ % classname
    except :
        pass
    cls.__mutable__ = __mutable__
    def hashed (self) :
        "Return 'True' if the %s has been hashed, 'False' otherwise."
        return hasattr(self, "_hash")
    try :
        hashed.__doc__ = hashed.__doc__ % classname
    except :
        pass
    cls.hashed = hashed
    def mutable (self) :
        "Return 'True' if the %s is not hashed, 'False' otherwise."
        return not self.hashed()
    try :
        mutable.__doc__ = mutable.__doc__ % classname
    except :
        pass
    cls.mutable = mutable
    return cls

class hlist (list) :
    """Hashable lists. They support all standard methods from `list`.

    >>> l = hlist(range(5))
    >>> l
    hlist([0, 1, 2, 3, 4])
    >>> l.append(5)
    >>> l
    hlist([0, 1, 2, 3, 4, 5])
    >>> _ = hash(l)
    >>> l.append(6)
    Traceback (most recent call last):
    ...
    ValueError: hashed 'hlist' object is not mutable
    >>> unhash(l)
    >>> l.append(6)
    """
    # apidoc stop
    def __add__ (self, other) :
        return self.__class__(list.__add__(self, other))
    def __delitem__ (self, item) :
        self.__mutable__()
        list.__delitem__(self, item)
    def __delslice__ (self, *l, **d) :
        self.__mutable__()
        list.__delslice__(self, *l, **d)
    def __getslice__ (self, first, last) :
        return self.__class__(list.__getslice__(self, first, last))
    def __getitem__ (self, item) :
        ret = list.__getitem__(self, item)
        if ret.__class__ is list :
            return self.__class__(ret)
        return ret
    def __hash__ (self) :
        return hash(tuple(self))
    def __iadd__ (self, other) :
        return self.__class__(list.__iadd__(self, other))
    def __imul__ (self, n) :
        return self.__class__(list.__imul__(self, n))
    def __mul__ (self, n) :
        return self.__class__(list.__mul__(self, n))
    def __repr__ (self) :
        """
        >>> repr(hlist(range(3)))
        'hlist([0, 1, 2])'
        """
        return "%s(%s)" % (self.__class__.__name__, list.__repr__(self))
    def __rmul__ (self, n) :
        return self._class__(list.__rmul__(self, n))
    def __setitem__ (self, index, item) :
        self.__mutable__()
        list.__setitem__(self, index, item)
    def __setslice__ (self, first, last, item) :
        self.__mutable__()
        list.__setslice__(self, first, last, item)
    def append (self, item) :
        self.__mutable__()
        list.append(self, item)
    def extend (self, iterable) :
        self.__mutable__()
        list.extend(self, iterable)
    def insert (self, index, item) :
        self.__mutable__()
        list.insert(self, index, item)
    def pop (self, index=-1) :
        self.__mutable__()
        return list.pop(self, index)
    def remove (self, item) :
        self.__mutable__()
        list.remove(self, item)
    def reverse (self) :
        self.__mutable__()
        list.reverse(self)
    def sort (self, cmp=None, key=None, reverse=False) :
        self.__mutable__()
        list.sort(self, cmp, key, reverse)

hlist = hashable(hlist)

class hdict (dict) :
    """Hashable dictionnaries. They support all standard methods from
    `dict`.

    >>> l = hlist(range(5))
    >>> d = hdict([(l, 0)])
    >>> d
    hdict({hlist([0, 1, 2, 3, 4]): 0})
    >>> l in d
    True
    >>> [0, 1, 2, 3, 4] in d
    Traceback (most recent call last):
    ...
    TypeError: ...
    >>> hlist([0, 1, 2, 3, 4]) in d
    True
    >>> d[hlist([0, 1, 2, 3, 4])]
    0
    >>> l.append(5)
    Traceback (most recent call last):
    ...
    ValueError: hashed 'hlist' object is not mutable
    >>> _ = hash(d)
    >>> d.pop(l)  # any mutation would produce the same error
    Traceback (most recent call last):
    ...
    ValueError: hashed 'hdict' object is not mutable
    >>> unhash(d)
    >>> d.pop(l)
    0
    """
    # apidoc stop
    def __delitem__ (self, key) :
        self.__mutable__()
        dict.__delitem__(self, key)
    def __hash__ (self) :
        """
        >>> _ = hash(hdict(a=1, b=2))
        """
        # 252756382 = hash("snakes.hashables.hlist")
        return reduce(xor, (hash(i) for i in self.items()), 252756382)
    def __repr__ (self) :
        """
        >>> repr(hdict(a=1))
        "hdict({'a': 1})"
        """
        return "%s(%s)" % (self.__class__.__name__, dict.__repr__(self))
    def __setitem__ (self, key, item) :
        self.__mutable__()
        dict.__setitem__(self, key, item)
    def clear (self) :
        self.__mutable__()
        dict.clear(self)
    def copy (self) :
        return self.__class__(dict.copy(self))
    @classmethod
    def fromkeys (_class, *args) :
        return _class(dict.fromkeys(*args))
    def pop (self, *args) :
        self.__mutable__()
        return dict.pop(self, *args)
    def popitem (self) :
        self.__mutable__()
        return dict.popitem(self)
    def setdefault (self, key, item=None) :
        self.__mutable__()
        return dict.setdefault (self, key, item)
    def update (self, other, **more) :
        self.__mutable__()
        return dict.update(self, other, **more)

hdict = hashable(hdict)

class hset (set) :
    """Hashable sets. They support all standard methods from `set`.

    >>> s = hset()
    >>> l = hlist(range(5))
    >>> s.add(l)
    >>> s
    hset([hlist([0, 1, 2, 3, 4])])
    >>> l in s
    True
    >>> [0, 1, 2, 3, 4] in s
    Traceback (most recent call last):
    ...
    TypeError: ...
    >>> hlist([0, 1, 2, 3, 4]) in s
    True
    >>> l.append(5)
    Traceback (most recent call last):
    ...
    ValueError: hashed 'hlist' object is not mutable
    >>> _ = hash(s)
    >>> s.discard(l)  # any mutation would produce the same error
    Traceback (most recent call last):
    ...
    ValueError: hashed 'hset' object is not mutable
    >>> unhash(s)
    >>> s.discard(l)
    """
    # apidoc stop
    def __and__ (self, other) :
        return self.__class__(set.__and__(self, other))
    def __hash__ (self) :
        """
        >>> _ = hash(hset([1, 2, 3]))
        >>> _ = hash(hset(range(5)) - set([0, 4]))
        """
        # 196496309 = hash("snakes.hashables.hset")
        return reduce(xor, (hash(x) for x in self), 196496309)
    def __iand__ (self, other) :
        return self.__class__(set.__iand__(self, other))
    def __ior__ (self, other) :
        return self.__class__(set.__ior__(self, other))
    def __isub__ (self, other) :
        return self.__class__(set.__isub__(self, other))
    def __ixor__ (self, other) :
        return self.__class__(set.__ixor__(self, other))
    def __or__ (self, other) :
        return self.__class__(set.__or__(self, other))
    def __rand__ (self, other) :
        return self.__class__(set.__rand__(self, other))
    def __repr__ (self) :
        """
        >>> repr(hset([1]))
        'hset([1])'
        """
        return "%s([%s])" % (self.__class__.__name__,
                             set.__repr__(self)[6:-2])
    def __ror__ (self, other) :
        return self.__class__(set.__ror__(self, other))
    def __rsub__ (self, other) :
        return self.__class__(set.__rsub__(self, other))
    def __rxor__ (self, other) :
        return self.__class__(set.__rxor__(self, other))
    def __str__ (self) :
        return self.__class__.__name__ + "(" + set.__str__(self).split("(", 1)[1]
    def __sub__ (self, other) :
        return self.__class__(set.__sub__(self, other))
    def __xor__ (self, other) :
        return self.__class__(set.__xor__(self, other))
    def add (self, item) :
        self.__mutable__()
        set.add(self, item)
    def clear (self) :
        self.__mutable__()
        set.clear(self)
    def copy (self) :
        return self.__class__(set.copy(self))
    def difference (self, other) :
        return self.__class__(set.difference(self, other))
    def difference_update (self, other) :
        self.__mutable__()
        set.difference_update(self, other)
    def discard (self, item) :
        self.__mutable__()
        set.discard(self, item)
    def intersection (self, other) :
        return self._class__(set.intersection(self, other))
    def intersection_update (self, other) :
        self.__mutable__()
        set.intersection_update(self, other)
    def pop (self) :
        self.__mutable__()
        return set.pop(self)
    def remove (self, item) :
        self.__mutable__()
        set.remove(self, item)
    def symmetric_difference (self, other) :
        self.__mutable__()
        set.symmetric_difference(self, other)
    def symmetric_difference_update (self, other) :
        self.__mutable__()
        set.symmetric_difference_update(self, other)
    def union (self, other) :
        return self.__class__(set.union(self, other))
    def update (self, other) :
        self.__mutable__()
        set.update(self, other)

hset = hashable(hset)



import glob, os, os.path

for src in glob.glob("snakes/lang/*/*.pgen") :
    tgt = os.path.join(os.path.dirname(src), "pgen.py")
    if not os.path.isfile(tgt) or os.path.getmtime(src) > os.path.getmtime(tgt) :
        print("python snakes/lang/pgen.py --output=%s %s" % (tgt, src))
        os.system("python snakes/lang/pgen.py --output=%s %s" % (tgt, src))

for src in glob.glob("snakes/lang/*/*.asdl") :
    tgt = os.path.join(os.path.dirname(src), "asdl.py")
    if not os.path.isfile(tgt) or os.path.getmtime(src) > os.path.getmtime(tgt) :
        print("python snakes/lang/asdl.py --output=%s %s" % (tgt, src))
        os.system("python snakes/lang/asdl.py --output=%s %s" % (tgt, src))



import sys
import snakes.plugins

from collections import defaultdict
from snakes.nets import ConstraintError
from snakes.typing import Instance, CrossProduct, tNatural
from snakes.data import iterate, WordSet

# apidoc skip
class Pid (tuple) :
    def __new__(cls, *args):
        """
        >>> Pid(1, 2, 3)
        Pid(1, 2, 3)
        >>> Pid('1.2.3')
        Pid(1, 2, 3)
        >>> Pid('1.2', 3)
        Pid(1, 2, 3)
        >>> Pid([1, 2], [3])
        Pid(1, 2, 3)
        >>> Pid((1, 2), 3)
        Pid(1, 2, 3)
        """
        data = []
        for a in args :
            if isinstance(a, int) :
                data.append(a)
            elif isinstance(a, str) :
                data.extend(int(x) for x in a.strip(".").split("."))
            elif isinstance(a, cls) :
                data.extend(a)
            elif isinstance(a, (list, tuple)) :
                data.extend(int(x) for x in a)
            else :
                raise ValueError("invalid Pid fragment %r" % a)
        return super(Pid, cls).__new__(cls, data)
    def __repr__(self) :
        """
        >>> Pid(1, 2, 3)
        Pid(1, 2, 3)
        >>>
        """
        return "Pid(%s)" % ", ".join(str(x) for x in self)
    def __str__ (self) :
        """
        >>> str(Pid(1, 2, 3))
        '1.2.3'
        """
        return ".".join(str(x) for x in self)
    @classmethod
    def from_str (cls, data) :
        """
        >>> Pid.from_str('1.2.3')
        Pid(1, 2, 3)
        """
        return cls(data)
    @classmethod
    def from_list (cls, data) :
        """
        >>> Pid.from_str([1, 2, 3])
        Pid(1, 2, 3)
        """
        return cls(data)
    def copy (self) :
        """
        >>> p = Pid(1, 2, 3)
        >>> p.copy()
        Pid(1, 2, 3)
        >>> p == p.copy()
        True
        >>> p is p.copy()
        False
        """
        return self.__class__(*self)
    def __add__(self, frag):
        """
        >>> Pid(1, 2, 3) + 4
        Pid(1, 2, 3, 4)
        >>> Pid(1, 2, 3) + Pid(4, 5)
        Pid(1, 2, 3, 4, 5)
        >>> Pid(1, 2, 3) + [4, 5]
        Pid(1, 2, 3, 4, 5)
        """
        return self.__class__(self, frag)
    def at(self, i):
        """
        >>> Pid(1, 2, 3).at(1)
        2
        """
        return self[i]
    def __getitem__ (self, start, stop=None) :
        if stop is None :
            return tuple.__getitem__(self, start)
        else :
            return self.__getslice__(start, stop)
    def __getslice__ (self, start, stop=sys.maxint) :
        """
        >>> pid = Pid(1, 2, 3, 4)
        >>> pid[:]
        Pid(1, 2, 3, 4)
        >>> pid[1:]
        Pid(2, 3, 4)
        >>> pid[:-1]
        Pid(1, 2, 3)
        >>> pid[1:-1]
        Pid(2, 3)
        >>> pid[1:3]
        Pid(2, 3)
        """
        return self.__class__(tuple.__getslice__(self, start, stop))
    def subpid (self, start=0, stop=sys.maxint) :
        """
        >>> pid = Pid(1, 2, 3, 4)
        >>> pid.subpid(0)
        Pid(1, 2, 3, 4)
        >>> pid.subpid(1)
        Pid(2, 3, 4)
        >>> pid.subpid(0, -1)
        Pid(1, 2, 3)
        >>> pid.subpid(1, -1)
        Pid(2, 3)
        >>> pid.subpid(1, 3)
        Pid(2, 3)
        """
        return self[start:stop]
    def prefix(self):
        """
        >>> Pid(1, 2, 3).prefix()
        Pid(1, 2)
        """
        return self[:-1]
    def suffix(self):
        """
        >>> Pid(1, 2, 3).suffix()
        Pid(2, 3)
        """
        return self[1:]
    def ends_with(self):
        """
        >>> Pid(1, 2, 3).ends_with()
        3
        """
        return self[-1]
    def next(self, pid_component):
        """
        >>> Pid(1).next(0)
        Pid(1, 1)
        >>> Pid(1, 2).next('2').next(3)
        Pid(1, 2, 3, 4)
        """
        return self + [int(pid_component)+1]
    def parent(self, other):
        """
        >>> Pid(1, 1, 1).parent(Pid(1, 1, 1))
        False
        >>> Pid(1, 1, 1).parent(Pid(1, 1, 1, 3))
        True
        >>> Pid(1, 1, 1).parent(Pid(1, 1))
        False
        >>> Pid(1, 1, 1).parent(Pid(1, 1, 1, 2, 4))
        True
        """
        return len(self) < len(other) and other[:len(self)] == self
    def parent1(self, other):
        """
        >>> Pid(1, 1, 1).parent1(Pid(1, 1, 1))
        False
        >>> Pid(1, 1, 1).parent1(Pid(1, 1, 1, 3))
        True
        >>> Pid(1, 1, 1).parent1(Pid(1, 1))
        False
        >>> Pid(1, 1, 1).parent1(Pid(1, 1, 1, 2, 4))
        False
        """
        return len(self)+1 == len(other) and other[:-1] == self
    def sibling(self, other):
        """
        >>> Pid(1, 1, 1).sibling(Pid(1, 1, 1))
        False
        >>> Pid(1, 1, 1).sibling(Pid(1, 1, 1, 3))
        False
        >>> Pid(1, 1, 1).sibling(Pid(1, 1, 2))
        True
        >>> Pid(1, 1, 2).sibling(Pid(1, 1, 1))
        False
        >>> Pid(1, 1, 1).sibling(Pid(1, 1, 5))
        True
        """
        return self[:-1] == other[:-1] and self[-1] < other[-1]
    def sibling1(self, other):
        """
        >>> Pid(1, 1, 1).sibling1(Pid(1, 1, 1))
        False
        >>> Pid(1, 1, 1).sibling1(Pid(1, 1, 1, 3))
        False
        >>> Pid(1, 1, 1).sibling1(Pid(1, 1, 2))
        True
        >>> Pid(1, 1, 2).sibling1(Pid(1, 1, 1))
        False
        >>> Pid(1, 1, 1).sibling1(Pid(1, 1, 5))
        False
        """
        return self[:-1] == other[:-1] and self[-1]+1 == other[-1]

tPid = snakes.typing.Instance(Pid)
tNextPid = CrossProduct(tPid, tNatural)

# apidoc skip
class ChildPid (object) :
    def __init__ (self, parent) :
        self.parent = parent

# apidoc skip
class ParentPid (object) :
    def __init__ (self, name, env) :
        self.name = name
        self.env = env
        self.count = 0
    def kill (self) :
        self.env.killed.add(self.name)
    def new (self) :
        child = ChildPid(self)
        self.count += 1
        return child

# apidoc skip
class PidEnv (dict) :
    def __init__ (self, prog) :
        dict.__init__(self)
        self.killed = set()
        self.spawned = defaultdict(list)
        self.next = {}
        self._vars = set()
        self(prog)
    def __call__ (self, prog) :
        exec(prog, self)
        self._vars.update(self.killed)
        for l in self.spawned.values() :
            self._vars.update(l)
    def __getitem__ (self, name) :
        if name in self :
            return dict.__getitem__(self, name)
        else :
            pid = self[name] = ParentPid(name, self)
            return pid
    def __setitem__ (self, name, value) :
        if isinstance(value, ChildPid) :
            self.spawned[value.parent.name].append(name)
        dict.__setitem__(self, name, value)
    def vars (self) :
        return set(self._vars)

@snakes.plugins.plugin("snakes.nets",
                       depends=["snakes.plugins.let",
                                "snakes.plugins.status"])
def extend (module) :
    snk = module
    class Transition (snk.Transition) :
        def __init__ (self, name, guard=None, **args) :
            self.pids = PidEnv(args.pop("pids", ""))
            vars = WordSet(self.pids.vars())
            if self.pids.spawned :
                assign = []
                for parent, children in self.pids.spawned.items() :
                    pidcount = vars.fresh(add=True, base="next_%s" % parent)
                    self.pids.next[parent] = pidcount
                    for n, child in enumerate(children) :
                        assign.append("%s=%s.next(%s+%s)"
                                      % (child, parent, n, pidcount))
                if guard is None :
                    guard = snk.Expression("newpids(%s)" % ", ".join(assign))
                else :
                    guard = guard & snk.Expression("newpids(%s)"
                                                   % ", ".join(assign))
            snk.Transition.__init__(self, name, guard, **args)
        def vars (self) :
            return self.pids.vars() | snk.Transition.vars(self)
    class PetriNet (snk.PetriNet) :
        def __init__ (self, name, **args) :
            snk.PetriNet.__init__(self, name, **args)
            self.globals["newpids"] = self.globals["let"]
            self.nextpids, nextpids = None, args.pop("nextpids", "nextpids")
            self.add_place(snk.Place(nextpids, [], tNextPid,
                                     status=snk.buffer(nextpids)))
            self.nextpids = nextpids
        def copy (self, name=None) :
            if name is None :
                name = self.name
            result = self.__class__(name)
            result._declare = self._declare[:]
            result.globals = self.globals.copy()
            for place in self.place() :
                if place.name != self.nextpids :
                    result.add_place(place.copy())
            for trans in self.transition() :
                result.add_transition(trans.copy())
                for place, label in trans.input() :
                    result.add_input(place.name, trans.name, label.copy())
                for place, label in trans.output() :
                    result.add_output(place.name, trans.name, label.copy())
            return result
        def add_place (self, place, **args) :
            if place.name == self.nextpids :
                raise ConstraintError("reserved place name %r"
                                      % self.nextpids)
            snk.PetriNet.add_place(self, place, **args)
        def add_transition (self, trans, **args) :
            snk.PetriNet.add_transition(self, trans, **args)
            cons, prod = {}, {}
            for parent, children in trans.pids.spawned.items() :
                pidcount = trans.pids.next[parent]
                cons[parent] = snk.Tuple([snk.Variable(parent),
                                          snk.Variable(pidcount)])
                prod[parent] = snk.Tuple([snk.Variable(parent),
                                          snk.Expression("%s+%s" %
                                                         (pidcount,
                                                          len(children)))])
                for num, child in enumerate(children) :
                    prod[child] = snk.Tuple([snk.Variable(child),
                                             snk.Value(0)])
            for pid in trans.pids.killed :
                prod.pop(pid, None)
            if len(cons) > 1 :
                self.add_input(self.nextpids, trans.name,
                               snk.MultiArc(cons.values()))
            elif len(cons) == 1 :
                self.add_input(self.nextpids, trans.name,
                               iter(cons.values()).next())
            if len(prod) > 1 :
                self.add_output(self.nextpids, trans.name,
                                snk.MultiArc(prod.values()))
            elif len(cons) == 1 :
                self.add_output(self.nextpids, trans.name,
                                iter(prod.values()).next())
    return PetriNet, Transition, Pid, ("tPid", tPid), ("tNextPid", tNextPid)



"""SNAKES library is organised into three parts:

  * the core library is package `snakes` and its modules, among which
    `snakes.nets` is the one to work with Petri nets while the others
    can be seen as auxiliary modules
  * the plugin system and the plugins themselves reside into package
    `snakes.plugins`
  * auxiliary tools are kept into other sub-packages: `snakes.lang`
    for all the material related to parsing Python and other
    languages, `snakes.utils` for various utilities like the ABCD
    compiler

@author: Franck Pommereau
@organization: University of Evry/Paris-Saclay
@copyright: (C) 2005-2013 Franck Pommereau
@license: GNU Lesser General Public Licence (aka. GNU LGPL), see the
    file `doc/COPYING` in the distribution or visit the [GNU web
    site](http://www.gnu.org/licenses/licenses.html#LGPL)
@contact: franck.pommereau@ibisc.univ-evry.fr
"""

version = "0.9.19"
defaultencoding = "utf-8"

"""## Module `snakes`

This module only provides the exceptions used throughout SNAKES.
"""

class SnakesError (Exception) :
    "Generic error in SNAKES"
    pass

class ConstraintError (SnakesError) :
    "Violation of a constraint"
    pass

class NodeError (SnakesError) :
    "Error related to a place or a transition"
    pass

class DomainError (SnakesError) :
    "Function applied out of its domain"
    pass

class ModeError (SnakesError) :
    "The modes of a transition cannot be found"
    pass

class PluginError (SnakesError) :
    "Error when adding a plugin"
    pass

class UnificationError (SnakesError) :
    "Error while unifying parameters"
    pass



"""Basic data types and related functions used in SNAKES
"""

import operator, inspect
from snakes.compat import *
from snakes import DomainError
from snakes.hashables import hdict
from snakes.pnml import Tree

def cross (sets) :
    """Cross-product of some iterable collections (typically, `list`
    or `set`).

    >>> list(cross([[1, 2], [3, 4, 5]]))
    [(1, 3), (1, 4), (1, 5), (2, 3), (2, 4), (2, 5)]
    >>> list(cross([[1, 2], [3, 4, 5], [6, 7, 8, 9]]))
    [(1, 3, 6), (1, 3, 7), (1, 3, 8), (1, 3, 9), (1, 4, 6), (1, 4, 7),
     (1, 4, 8), (1, 4, 9), (1, 5, 6), (1, 5, 7), (1, 5, 8), (1, 5, 9),
     (2, 3, 6), (2, 3, 7), (2, 3, 8), (2, 3, 9), (2, 4, 6), (2, 4, 7),
     (2, 4, 8), (2, 4, 9), (2, 5, 6), (2, 5, 7), (2, 5, 8), (2, 5, 9)]
    >>> list(cross([[], [1]]))
    []

    @param sets: the sets of values to use
    @type sets: `iterable`
    @return: an iterator over the tuples in the cross-product
    @rtype: `generator`
    """
    if len(sets) == 0 :
        pass
    elif len(sets) == 1 :
        for item in sets[0] :
            yield (item,)
    else :
        for item in sets[0] :
            for others in cross(sets[1:]) :
                yield (item,) + others

def iterate (value) :
    """Like Python's builtin `iter` but consider strings as atomic.

    >>> list(iter([1, 2, 3]))
    [1, 2, 3]
    >>> list(iterate([1, 2, 3]))
    [1, 2, 3]
    >>> list(iter('foo'))
    ['f', 'o', 'o']
    >>> list(iterate('foo'))
    ['foo']

    @param value: any object
    @type value: `object`
    @return: an iterator on the elements of `value` if is is iterable
        and is not string, an iterator on the sole `value` otherwise
    @rtype: `generator`
    """
    if isinstance(value, str) :
        return iter([value])
    else :
        try :
            return iter(value)
        except TypeError :
            return iter([value])

class WordSet (set) :
    """A set of words being able to generate fresh words.
    """
    def fresh (self, add=False, min=1, base="",
               allowed="abcdefghijklmnopqrstuvwxyz") :
        """Create a fresh word (ie, which is not in the set).

        >>> w = WordSet(['foo', 'bar'])
        >>> list(sorted(w))
        ['bar', 'foo']
        >>> w.fresh(True, 3)
        'aaa'
        >>> list(sorted(w))
        ['aaa', 'bar', 'foo']
        >>> w.fresh(True, 3)
        'baa'
        >>> list(sorted(w))
        ['aaa', 'baa', 'bar', 'foo']

        @param add: add the created word to the set if `add=True`
        @type add: `bool`
        @param min: minimal length of the new word
        @type min: `int`
        @param allowed: characters allowed in the new word
        @type allowed: `str`
        @param base: prefix of generated words
        @type base: `str`
        """
        if base :
            result = [base] + [allowed[0]] * max(0, min - len(base))
            if base in self :
                result.append(allowed[0])
                pos = len(result) - 1
            elif len(base) < min :
                pos = 1
            else :
                pos = 0
        else :
            result = [allowed[0]] * min
            pos = 0
        while "".join(result) in self :
            for c in allowed :
                try :
                    result[pos] = c
                except IndexError :
                    result.append(c)
                if "".join(result) not in self :
                    break
            pos += 1
        if add :
            self.add("".join(result))
        return "".join(result)

class MultiSet (hdict) :
    """Set with repetitions, ie, function from values to integers.

    MultiSets support various operations, in particular: addition
    (`+`), substraction (`-`), multiplication by a non negative
    integer (`*k`), comparisons (`<`, `>`, etc.), length (`len`).
    """
    def __init__ (self, values=[]) :
        """Initialise the multiset, adding values to it.

        >>> MultiSet([1, 2, 3, 1, 2])
        MultiSet([...])
        >>> MultiSet()
        MultiSet([])

        @param values: a single value or an iterable collection of
            values (strings are not iterated)
        @type values: `object`
        """
        self.add(values)
    def copy (self) :
        """Copy a `MultiSet`

        >>> m1 = MultiSet([1, 2, 3, 1, 2])
        >>> m2 = m1.copy()
        >>> m1 == m2 and m1 is not m2
        True

        @return: a copy of the multiset
        @rtype: `MultiSet`
        """
        result = MultiSet()
        result.update(self)
        return result
    __pnmltag__ = "multiset"
    # apidoc skip
    def __pnmldump__ (self) :
        """
        >>> MultiSet([1, 2, 3, 4, 1, 2]).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <multiset>
          <item>
           <value>
            <object type="int">
            ...
            </object>
           </value>
           <multiplicity>
           ...
           </multiplicity>
          </item>
          <item>
           <value>
            <object type="int">
            ...
            </object>
           </value>
           <multiplicity>
           ...
           </multiplicity>
          </item>
          <item>
           <value>
            <object type="int">
            ...
            </object>
           </value>
           <multiplicity>
           ...
           </multiplicity>
          </item>
          <item>
           <value>
            <object type="int">
            ...
            </object>
           </value>
           <multiplicity>
           ...
           </multiplicity>
          </item>
         </multiset>
        </pnml>
        """
        nodes = []
        for value in hdict.__iter__(self) :
            nodes.append(Tree("item", None,
                              Tree("value", None, Tree.from_obj(value)),
                              Tree("multiplicity", str(self[value]))))
        return Tree(self.__pnmltag__, None, *nodes)
    # apidoc skip
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Load a multiset from its PNML representation

        >>> t = MultiSet([1, 2, 3, 4, 1, 2]).__pnmldump__()
        >>> MultiSet.__pnmlload__(t)
        MultiSet([...])
        """
        result = cls()
        for item in tree :
            times = int(item.child("multiplicity").data)
            value = item.child("value").child().to_obj()
            result._add(value, times)
        return result
    def _add (self, value, times=1) :
        """Add a single value `times` times.

        @param value: the value to add
        @type value: `object`
        @param times: the number of times that `value` has to be added
        @type times: non negative `int`
        @raise ValueError: when `times` is negative
        """
        if times < 0 :
            raise ValueError("negative values are forbidden")
        try :
            self[value] += times
        except KeyError :
            self[value] = times
    def add (self, values, times=1) :
        """Add values to the multiset.

        >>> m = MultiSet()
        >>> m.add([1, 2, 2, 3], 2)
        >>> list(sorted(m.items()))
        [1, 1, 2, 2, 2, 2, 3, 3]
        >>> m.add(5, 3)
        >>> list(sorted(m.items()))
        [1, 1, 2, 2, 2, 2, 3, 3, 5, 5, 5]

        @param values: the values to add or a single value to add
        @type values: `object`
        @param times: the number of times each value should be added
            (must be non-negative)
        @type times: `int`
        @raise ValueError: when `times` is negative
        """
        self.__mutable__()
        for value in iterate(values) :
            self._add(value, times)
    def _remove (self, value, times=1) :
        """Remove a single value `times` times.

        @param value: the value to remove
        @type value: any object
        @param times: the number of times that `value` has to be
            removed
        @type times: non negative `int`
        @raise ValueError: when `times` is negative
        """
        if times < 0 :
            raise ValueError("negative values are forbidden")
        if times > self.get(value, 0) :
            raise ValueError("not enough occurrences")
        self[value] -= times
        if self[value] == 0 :
            del self[value]
    def remove (self, values, times=1) :
        """Remove values to the multiset.

        >>> m = MultiSet([1, 2, 2, 3] * 2)
        >>> list(sorted(m.items()))
        [1, 1, 2, 2, 2, 2, 3, 3]
        >>> m.remove(2, 3)
        >>> list(sorted(m.items()))
        [1, 1, 2, 3, 3]
        >>> m.remove([1, 3], 2)
        >>> list(sorted(m.items()))
        [2]

        @param values: the values to remove or a single value to
            remove
        @type values: `object`
        @param times: the number of times each value should be removed
            (must be non negative)
        @type times: `int`
        @raise ValueError: when `times` is negative
        """
        self.__mutable__()
        for value in iterate(values) :
            self._remove(value, times)
    def __call__ (self, value) :
        """Number of occurrences of `value`.

        >>> m = MultiSet([1, 1, 2, 3, 3, 3])
        >>> m(1), m(2), m(3), m(4)
        (2, 1, 3, 0)

        @param value: the value the count
        @type value: `object`
        @rtype: `int`
        """
        return self.get(value, 0)
    def __iter__ (self) :
        """Iterate over the values, _including repetitions_. Use
        `MultiSet.keys` to ignore repetitions.

        >>> list(sorted(iter(MultiSet([1, 2, 3, 1, 2]))))
        [1, 1, 2, 2, 3]

        @return: an iterator on the elements
        @rtype: `generator`
        """
        for value in dict.__iter__(self) :
            for count in range(self[value]) :
                yield value
    def items (self) :
        """Return the list of items with repetitions. The list without
        repetitions can be retrieved with `MultiSet.key`.

        >>> m = MultiSet([1, 2, 2, 3])
        >>> list(sorted(m.items()))
        [1, 2, 2, 3]
        >>> list(sorted(m.keys()))
        [1, 2, 3]

        @return: list of items including repetitions
        @rtype: `list`
        """
        return list(iter(self))
    def __str__ (self) :
        """Return a simple string representation of the multiset

        >>> str(MultiSet([1, 2, 2, 3]))
        '{...}'

        @return: simple string representation of the multiset
        @rtype: `str`
        """
        return "{%s}" % ", ".join(repr(x) for x in self)
    def __repr__ (self) :
        """Return a string representation of the multiset that is
        suitable for `eval`

        >>> repr(MultiSet([1, 2, 2, 3]))
        'MultiSet([...])'

        @return: precise string representation of the multiset
        @rtype: `str`
        """
        return "MultiSet([%s])" % ", ".join(repr(x) for x in self)
    def __len__ (self) :
        """Return the number of elements, _including repetitions_.

        >>> len(MultiSet([1, 2] * 3))
        6

        @rtype: `int`
        """
        if self.size() == 0 :
            return 0
        else :
            return reduce(operator.add, self.values())
    def size (self) :
        """Return the number of elements, _excluding repetitions_.

        >>> MultiSet([1, 2] * 3).size()
        2

        @rtype: `int`
        """
        return dict.__len__(self)
    def __add__ (self, other) :
        """Adds two multisets.

        >>> MultiSet([1, 2, 3]) + MultiSet([2, 3, 4])
        MultiSet([...])

        @param other: the multiset to add
        @type other: `MultiSet`
        @rtype: `MultiSet`
        """
        result = self.copy()
        for value, times in dict.items(other) :
            result._add(value, times)
        return result
    def __sub__ (self, other) :
        """Substract two multisets. The second multiset must be
        smaller than the first one.

        >>> MultiSet([1, 2, 3]) - MultiSet([2, 3])
        MultiSet([1])
        >>> MultiSet([1, 2, 3]) - MultiSet([2, 3, 4])
        Traceback (most recent call last):
        ...
        ValueError: not enough occurrences

        @param other: the multiset to substract
        @type other: `MultiSet`
        @rtype: `MultiSet`
        @raise ValueError: when the second multiset is not smaller
        """
        result = self.copy()
        for value, times in dict.items(other) :
            result._remove(value, times)
        return result
    def __mul__ (self, other) :
        """Multiplication by a non-negative integer.

        >>> MultiSet([1, 2]) * 3 == MultiSet([1, 2] * 3)
        True

        @param other: the integer to multiply
        @type other: non-negative `int`
        @rtype: `MultiSet`
        @raise ValueError: when `other` is negative
        """
        if other < 0 :
            raise ValueError("negative values are forbidden")
        elif other == 0 :
            return MultiSet()
        else :
            result = self.copy()
            for value in self.keys() :
                result[value] *= other
            return result
    __hash__ = hdict.__hash__
    def __eq__ (self, other) :
        """Test for equality.

        >>> MultiSet([1, 2, 3]*2) == MultiSet([1, 2, 3]*2)
        True
        >>> MultiSet([1, 2, 3]) == MultiSet([1, 2, 3, 3])
        False

        @param other: the multiset to compare with
        @type other: `MultiSet`
        @rtype: `bool`
        """
        if len(self) != len(other) :
            return False
        else :
            for val in self :
                try :
                    if self[val] != other[val] :
                        return False
                except (KeyError, TypeError) :
                    return False
        return True
    def __ne__ (self, other) :
        """Test for difference.

        >>> MultiSet([1, 2, 3]*2) != MultiSet([1, 2, 3]*2)
        False
        >>> MultiSet([1, 2, 3]) != MultiSet([1, 2, 3, 3])
        True

        @param other: the multiset to compare with
        @type other: `MultiSet`
        @rtype: `bool`
        """
        return not(self == other)
    def __lt__ (self, other) :
        """Test for strict inclusion. A multiset `A` is strictly
        included in a multiset `B` iff every element in `A` is also in
        `B` but less repetitions `A` than in `B`.

        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2, 3, 4])
        True
        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2, 3, 3])
        True
        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2, 3]) < MultiSet([1, 2])
        False
        >>> MultiSet([1, 2, 2]) < MultiSet([1, 2, 3, 4])
        False

        @param other: the multiset to compare with
        @type other: `MultiSet`
        @rtype: `bool`
        """
        if not set(self.keys()) <= set(other.keys()) :
            return False
        result = False
        for value, times in dict.items(self) :
            count = other(value)
            if times > count :
                return False
            elif times < count :
                result = True
        return result or (dict.__len__(self) < dict.__len__(other))
    def __le__ (self, other) :
        """Test for inclusion.

        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2, 3, 4])
        True
        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2, 3, 3])
        True
        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3]) <= MultiSet([1, 2])
        False
        >>> MultiSet([1, 2, 2]) <= MultiSet([1, 2, 3, 4])
        False

        @param other: the multiset to compare with
        @type other: `MultiSet`
        @rtype: `bool`
        """
        if not set(self.keys()) <= set(other.keys()) :
            return False
        for value, times in dict.items(self) :
            count = other(value)
            if times > count :
                return False
        return True
    def __gt__ (self, other) :
        """Test for strict inclusion.

        >>> MultiSet([1, 2, 3, 4]) > MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3, 3]) > MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3]) > MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2]) > MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2, 3, 4]) > MultiSet([1, 2, 2])
        False

        @param other: the multiset to compare with
        @type other: `MultiSet`
        @rtype: `bool`
        """
        return other.__lt__(self)
    def __ge__ (self, other) :
        """Test for inclusion.

        >>> MultiSet([1, 2, 3, 4]) >= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3, 3]) >= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2, 3]) >= MultiSet([1, 2, 3])
        True
        >>> MultiSet([1, 2]) >= MultiSet([1, 2, 3])
        False
        >>> MultiSet([1, 2, 3, 4]) >= MultiSet([1, 2, 2])
        False

        @param other: the multiset to compare with
        @type other: `MultiSet`
        @rtype: `bool`
        """
        return other.__le__(self)
    def domain (self) :
        """Return the domain of the multiset, that is, the set of
        elements that occurr at least once in the multiset.

        >>> list(sorted((MultiSet([1, 2, 3, 4]) + MultiSet([1, 2, 3])).domain()))
        [1, 2, 3, 4]

        @return: the set of values in the domain
        @rtype: `set`
        """
        return set(self.keys())

class Substitution (object) :
    """Map names to values or names, equals the identity where not
    defined.

    Substitutions support the `+` operation (union with consistency
    check between the two operands) and the `*` operation which is the
    composition of functions (`(f*g)(x)` is `f(g(x))`).

    Several methods (eg, `image`) return lists instead of sets, this
    avoids the restriction of having only hashable values in a
    substitution image.
    """
    def __init__ (self, *largs, **dargs) :
        """Initialise using a dictionnary as a mapping.

        The expected arguments are any ones acceptables for
        initializing a dictionnary.

        >>> Substitution()
        Substitution()
        >>> Substitution(x=1, y=2)
        Substitution(...)
        >>> Substitution([('x', 1), ('y', 2)])
        Substitution(...)
        >>> Substitution({'x': 1, 'y': 2})
        Substitution(...)
        """
        self._dict = dict(*largs, **dargs)
    # apidoc skip
    def __hash__ (self) :
        """
        >>> hash(Substitution(x=1, y=2)) == hash(Substitution(y=2, x=1))
        True
        """
        # 153913524 = hash('snakes.data.Substitution')
        return reduce(operator.xor,
                      (hash(i) for i in self._dict.items()),
                      153913524)
    def __eq__ (self, other) :
        """Test for equality.

        >>> Substitution(x=1, y=2) == Substitution(y=2, x=1)
        True
        >>> Substitution(x=1, y=2) == Substitution(y=1, x=1)
        False
        """
        try :
            return self._dict == other._dict
        except :
            return False
    def __ne__ (self, other) :
        """Test for inequality.

        >>> Substitution(x=1, y=2) != Substitution(y=2, x=1)
        False
        >>> Substitution(x=1, y=2) != Substitution(y=1, x=1)
        True
        """
        return not self.__eq__(other)
    __pnmltag__ = "substitution"
    # apidoc skip
    def __pnmldump__ (self) :
        """Dumps a substitution to a PNML tree

        >>> Substitution(x=1, y=2).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <substitution>
          <item>
           <name>
           ...
           </name>
           <value>
            <object type="int">
            ...
            </object>
           </value>
          </item>
          <item>
           <name>
           ...
           </name>
           <value>
            <object type="int">
            ...
            </object>
           </value>
          </item>
         </substitution>
        </pnml>

        @return: PNML representation
        @rtype: `snakes.pnml.Tree`
        """
        nodes = []
        for name, value in self._dict.items() :
            nodes.append(Tree("item", None,
                              Tree("name", name),
                              Tree("value", None,
                                   Tree.from_obj(value))))
        return Tree(self.__pnmltag__, None, *nodes)
    # apidoc skip
    @classmethod
    def __pnmlload__ (cls, tree) :
        """Load a substitution from its PNML representation

        >>> t = Substitution(x=1, y=2).__pnmldump__()
        >>> Substitution.__pnmlload__(t)
        Substitution(...)

        @param tree: the PNML tree to load
        @type tree: `snakes.pnml.Tree`
        @return: the substitution loaded
        @rtype: `Substitution`
        """
        result = cls()
        for item in tree :
            name = item.child("name").data
            value = item.child("value").child().to_obj()
            result._dict[name] = value
        return result
    def items (self) :
        """Return the list of pairs `(name, value)` such that the
        substitution maps each `name` to the correspondign `value`.

        >>> Substitution(x=1, y=2).items()
        [('...', ...), ('...', ...)]

        @return: a list of pairs (name, value) for each mapped name
        @rtype: `list`
        """
        return list(self._dict.items())
    def domain (self) :
        """Return the set of mapped names.

        >>> list(sorted(Substitution(x=1, y=2).domain()))
        ['x', 'y']

        @return: the set of mapped names
        @rtype: `set`
        """
        return set(self._dict.keys())
    def image (self) :
        """Return the list of values associated to the names.

        >>> list(sorted(Substitution(x=1, y=2).image()))
        [1, 2]

        @return: the list of values associated to names
        @rtype: `list`
        """
        return list(self._dict.values())
    def __contains__ (self, name) :
        """Test if a name is mapped by the substitution.

        >>> 'x' in Substitution(x=1, y=2)
        True
        >>> 'z' in Substitution(x=1, y=2)
        False

        @param name: the name to test
        @type name: `str` (usually)
        @return: a Boolean indicating whether this name is in the
            domain or not
        @rtype: `bool`
        """
        return name in self._dict
    def __iter__ (self) :
        """Iterate over the mapped names.

        >>> list(sorted(iter(Substitution(x=1, y=2))))
        ['x', 'y']

        @return: an iterator over the domain of the substitution
        @rtype: `generator`
        """
        return iter(self._dict)
    def __str__ (self) :
        """Return a compact string representation.

        >>> str(Substitution(x=1, y=2))
        '{... -> ..., ... -> ...}'

        @return: a simple string representation
        @rtype: `str`
        """
        return "{%s}" % ", ".join(["%s -> %r" % (str(var), val)
                                   for var, val in self.items()])
    def __repr__ (self) :
        """Return a string representation suitable for `eval`.

        >>> repr(Substitution(x=1, y=2))
        'Substitution(...)'

        @return: a precise string representation
        @rtype: `str`
        """
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join(("%s=%s" % (str(var), repr(val))
                                      for var, val in self.items())))
    def dict (self) :
        """Return the mapping as a dictionnary.

        >>> Substitution(x=1, y=2).dict() == {'x': 1, 'y': 2}
        True

        @return: a dictionnary that does the same mapping as the
            substitution
        @rtype: `dict`
        """
        return self._dict.copy()
    def copy (self) :
        """Return a distinct copy of the mapping.

        >>> s1 = Substitution(x=1, y=2)
        >>> s2 = s1.copy()
        >>> s1 == s2 and s1 is not s2
        True

        @return: a copy of the substitution
        @rtype: `Substitution`
        """
        return Substitution(self.dict())
    def __setitem__ (self, var, value) :
        """Assign an entry to the substitution

        >>> s = Substitution()
        >>> s['x'] = 42
        >>> s
        Substitution(x=42)

        @param var: the name of the variable
        @type var: `str`
        @param value: the value to which `var` is bound
        @type value: `object`
        """
        self._dict[var] = value
    def __getitem__ (self, var) :
        """Return the mapped value.

        >>> s = Substitution(x=1, y=2)
        >>> s['x']
        1
        >>> try : s['z']
        ... except DomainError : print(sys.exc_info()[1])
        unbound variable 'z'

        @param var: the name of the variable
        @type var: `str` (usually)
        @return: the value that `var` maps to
        @rtype: `object`
        @raise DomainError: if `var` does not belong to the domain
        """
        try :
            return self._dict[var]
        except KeyError :
            raise DomainError("unbound variable '%s'" % var)
    def __call__ (self, var) :
        """Return the mapped value or `var` itself if it is not
        mapped.

        >>> s = Substitution(x=1, y=2)
        >>> s('x')
        1
        >>> s('z')
        'z'

        @param var: the name of the variable
        @type var: `str` (usually)
        @return: the value that `var` maps to or `var` itself if it
            does not belong to the domain
        @rtype: `object`
        """
        try :
            return self._dict[var]
        except KeyError :
            return var
    def __add__ (self, other) :
        """Add two substitution.
        Fails with `DomainError` if the two substitutions map a same
        name to different values.

        >>> s = Substitution(x=1, y=2) + Substitution(y=2, z=3)
        >>> s('x'), s('y'), s('z')
        (1, 2, 3)
        >>> try : Substitution(x=1, y=2) + Substitution(y=4, z=3)
        ... except DomainError : print(sys.exc_info()[1])
        conflict on 'y'

        @param other: another substitution
        @type other: `Substitution`
        @return: the union of the substitutions
        @rtype: `Substitution`
        @raise DomainError: when a name is inconsistently mapped
        """
        for var in self :
            if var in other and (self[var] != other[var]) :
                raise DomainError("conflict on '%s'" % var)
        s = self.__class__(self.dict())
        s._dict.update(other.dict())
        return s
    def __mul__ (self, other) :
        """Compose two substitutions.
        The composition of `f` and `g` is such that `(f*g)(x)` is
        `f(g(x))`.

        >>> f = Substitution(a=1, d=3, y=5)
        >>> g = Substitution(b='d', c=2, e=4, y=6)
        >>> h = f*g
        >>> h('a'), h('b'), h('c'), h('d'), h('e'), h('y'), h('x')
        (1, 3, 2, 3, 4, 6, 'x')

        @param other: another substitution
        @type other: `Substitution`
        @return: the composition of the substitutions
        @rtype: `Substitution`
        """
        res = self.copy()
        for var in other :
            res._dict[var] = self(other(var))
        return res
    def restrict (self, domain) :
        """Restrict the substitution to `domain`, ie remove all
        elements that are not in `domain`. Note that `domain` may
        include names that are not in the substitution, they are
        simply ignored.

        >>> s = Substitution(a=1, b=2, c=3, d=4).restrict(['a', 'b', 'z'])
        >>> list(sorted(s.domain()))
        ['a', 'b']

        @param domain: the new domain as a set/list/... of names
        @type domain: `iterable`
        @return: the restricted substitution
        @rtype: `Substitution`
        """
        result = self.copy()
        for name in result.domain() - set(domain) :
            result._dict.pop(name, None)
        return result

class Symbol (object) :
    """A symbol that may be used as a constant
    """
    def __init__ (self, name, export=True) :
        """If `export` is `True`, the created symbol is exported under
        its name. If `export` is `False`, no export is made. Finally,
        if `export` is a string, it specifies the name of the exported
        symbol. Exporting the name is made by adding it to the
        caller's global dict.

        @param name: the name (or value of the symbol)
        @type name: `str`
        @param export: the name under which the symbol is exported
        @type export: `str` or `bool` or `None`

        >>> Symbol('foo')
        Symbol('foo')
        >>> foo
        Symbol('foo')
        >>> Symbol('egg', 'spam')
        Symbol('egg', 'spam')
        >>> spam
        Symbol('egg', 'spam')
        >>> Symbol('bar', False)
        Symbol('bar', False)
        >>> bar
        Traceback (most recent call last):
         ...
        NameError: ...
        """
        self.name = name
        if export is True :
            export = name
        self._export = export
        if export :
            inspect.stack()[1][0].f_globals[export] = self
    __pnmltag__ = "symbol"
    # apidoc skip
    def __pnmldump__ (self) :
        """
        >>> Symbol('egg', 'spam').__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <symbol name="egg">
          <object type="str">
           spam
          </object>
         </symbol>
        </pnml>
        >>> Symbol('foo').__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <symbol name="foo"/>
        </pnml>
        >>> Symbol('bar', False).__pnmldump__()
        <?xml version="1.0" encoding="utf-8"?>
        <pnml>
         <symbol name="bar">
          <object type="bool">
           False
          </object>
         </symbol>
        </pnml>
        """
        if self.name == self._export :
            children = []
        else :
            children = [Tree.from_obj(self._export)]
        return Tree(self.__pnmltag__, None, *children, **dict(name=self.name))
    # apidoc skip
    @classmethod
    def __pnmlload__ (cls, tree) :
        """
        >>> Symbol.__pnmlload__(Symbol('foo', 'bar').__pnmldump__())
        Symbol('foo', 'bar')
        >>> Symbol.__pnmlload__(Symbol('foo').__pnmldump__())
        Symbol('foo')
        >>> Symbol.__pnmlload__(Symbol('foo', False).__pnmldump__())
        Symbol('foo', False)
        """
        name = tree["name"]
        try :
            export = tree.child().to_obj()
        except :
            export = name
        return cls(name, export)
    def __eq__ (self, other) :
        """Test for equality of two symbols, which is the equality of
        their names.

        >>> Symbol('foo', 'bar') == Symbol('foo')
        True
        >>> Symbol('egg') == Symbol('spam')
        False
        """
        try :
            return (self.__class__.__name__ == other.__class__.__name__
                    and self.name == other.name)
        except AttributeError :
            return False
    def __ne__ (self, other) :
        """Test for inequality.

        >>> Symbol('foo', 'bar') != Symbol('foo')
        False
        >>> Symbol('egg') != Symbol('spam')
        True
        """
        return not (self == other)
    # apidoc skip
    def __hash__ (self) :
        """
        >>> hash(Symbol('foo', 'bar')) == hash(Symbol('foo'))
        True
        """
        return hash((self.__class__.__name__, self.name))
    def __str__ (self) :
        """Short string representation

        >>> str(Symbol('foo'))
        'foo'
        """
        return self.name
    def __repr__ (self) :
        """String representation suitable for `eval`

        >>> Symbol('foo')
        Symbol('foo')
        >>> Symbol('egg', 'spam')
        Symbol('egg', 'spam')
        >>> Symbol('bar', False)
        Symbol('bar', False)
        """
        if self._export == self.name :
            return "%s(%r)" % (self.__class__.__name__, self.name)
        else :
            return "%s(%r, %r)" % (self.__class__.__name__, self.name,
                                   self._export)
